/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.Arrays;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import com.koutra.dist.proc.faucet.ReaderFaucet;
import com.koutra.dist.proc.faucet.StreamFaucet;
import com.koutra.dist.proc.faucet.XMLFaucet;
import com.koutra.dist.proc.pipeline.IdentityReaderToWriterPipelineItemTemplate;
import com.koutra.dist.proc.pipeline.IdentityStreamToStreamPipelineItemTemplate;
import com.koutra.dist.proc.pipeline.LineBasedReaderDemuxPipelineItem;
import com.koutra.dist.proc.pipeline.LineBasedReaderToWriterPipelineItemTemplate;
import com.koutra.dist.proc.pipeline.ReaderDemuxPipelineItem;
import com.koutra.dist.proc.pipeline.StreamDemuxPipelineItem;
import com.koutra.dist.proc.pipeline.XMLFilterDemuxPipelineItem;
import com.koutra.dist.proc.pipeline.XMLFilterPipelineItem;
import com.koutra.dist.proc.pipeline.XMLFilterPipelineItemTemplate;
import com.koutra.dist.proc.sink.StreamMuxSink;
import com.koutra.dist.proc.sink.StreamZipMuxSink;
import com.koutra.dist.proc.sink.WriterMuxSink;
import com.koutra.dist.proc.sink.XMLZipMuxSink;
import com.koutra.dist.proc.util.Util;
import com.koutra.dist.proc.xml.XPathXMLMuxFilter;

import junit.framework.TestCase;

public class MuxDmuxTest extends TestCase {

	public void testDemuxLineTransform() throws IOException {
		// Set up the transformation
		ReaderFaucet faucet = new ReaderFaucet("1", "test/greek_input.txt",
				"UTF-8");
		IdentityReaderToWriterPipelineItemTemplate transformerB =
			new IdentityReaderToWriterPipelineItemTemplate();
		LineBasedReaderDemuxPipelineItem transformerA = new LineBasedReaderDemuxPipelineItem(
				"2", transformerB) {
			private boolean changePipeline = false;

			@Override
			public boolean shouldChangePipeline() {
				boolean retVal = changePipeline;
				changePipeline = false;
				return retVal;
			}

			@Override
			public String transformLine(String line) {
				if (line == null || line.length() == 0)
					return "";
				changePipeline = true;
				return line + "\r\n";
			}
		};
		WriterMuxSink sink = new WriterMuxSink("4", "test/greek_output.txt",
				"Cp1253");

		// Hook everything up
		transformerA.hookupFaucet(faucet);
		transformerB.hookupFaucet(transformerA);
		transformerB.hookupSink(sink);

		// Start the ball rolling
		sink.consume();

		// Shut things down
		sink.dispose();

		// Verify the result
		Reader inputFileReader = new InputStreamReader(new FileInputStream(
				"test/greek_input.txt"), "UTF-8");
		Reader outputFileReader = new InputStreamReader(new FileInputStream(
				"test/greek_output.txt"), "Cp1253");
		Util.testReaderEquivalence(inputFileReader, outputFileReader);
		inputFileReader.close();
		outputFileReader.close();
		
		File fileToDelete = new File("test/greek_output.txt");
		fileToDelete.delete();
	}
	
	public static class IdentityLineBasedReaderToWriterTemplate
			extends LineBasedReaderToWriterPipelineItemTemplate {
		@Override
		public String transformLine(String line) {
			if (line == null)
				return "";
			return line + "\r\n";
		}
	}

	public void testLineBasedReaderToWriterTransform() throws IOException {
		// Set up the transformation
		ReaderFaucet faucet = new ReaderFaucet("1", "test/greek_input.txt",
				"UTF-8");
		IdentityLineBasedReaderToWriterTemplate transformerB =
			new IdentityLineBasedReaderToWriterTemplate();
		LineBasedReaderDemuxPipelineItem transformerA = new LineBasedReaderDemuxPipelineItem(
				"2", transformerB) {
			private boolean changePipeline = false;

			@Override
			public boolean shouldChangePipeline() {
				boolean retVal = changePipeline;
				changePipeline = false;
				return retVal;
			}

			@Override
			public String transformLine(String line) {
				if (line == null || line.length() == 0)
					return "";
				changePipeline = true;
				return line + "\r\n";
			}
		};
		WriterMuxSink sink = new WriterMuxSink("4", "test/greek_output.txt",
				"Cp1253");

		// Hook everything up
		transformerA.hookupFaucet(faucet);
		transformerB.hookupFaucet(transformerA);
		transformerB.hookupSink(sink);

		// Start the ball rolling
		sink.consume();

		// Shut things down
		sink.dispose();

		// Verify the result
		Reader inputFileReader = new InputStreamReader(new FileInputStream(
				"test/greek_input.txt"), "UTF-8");
		Reader outputFileReader = new InputStreamReader(new FileInputStream(
				"test/greek_output.txt"), "Cp1253");
		Util.testReaderEquivalence(inputFileReader, outputFileReader);
		inputFileReader.close();
		outputFileReader.close();
		
		File fileToDelete = new File("test/greek_output.txt");
		fileToDelete.delete();
	}

	public void testDemuxReaderToWriterTransform() throws IOException {
		// Set up the transformation
		ReaderFaucet faucet = new ReaderFaucet("1", "test/greek_input.txt",
				"UTF-8");
		IdentityReaderToWriterPipelineItemTemplate transformerB =
			new IdentityReaderToWriterPipelineItemTemplate();
		ReaderDemuxPipelineItem transformerA = new ReaderDemuxPipelineItem(
				"2", transformerB) {
			private char[] remainingBuffer;

			@Override
			public boolean shouldChangePipeline() {
				return remainingBuffer != null && remainingBuffer.length != 0;
			}
			
			@Override
			public char[] transformBuffer(char[] buffer, int off, int len) {
				if (remainingBuffer == null) {
					if (len == -1)
						return new char[0];
					
					if (len - off <= 50)
						return Arrays.copyOfRange(buffer, off, off + len);
					char[] retVal = Arrays.copyOfRange(buffer, off, off + 50);
					remainingBuffer = Arrays.copyOfRange(buffer, off + 50, off + len);
					return retVal;
				}
				
				char[] concatArray = new char[remainingBuffer.length + len - off];
				for (int i = 0; i < remainingBuffer.length; i++) {
					concatArray[i] = remainingBuffer[i];
				}
				for (int i = off; i < len; i++) {
					concatArray[remainingBuffer.length + i - off] = buffer[i];
				}
				
				if (concatArray.length <= 50) {
					remainingBuffer = null;
					return Arrays.copyOfRange(concatArray, 0, concatArray.length);
				}
				char[] retVal = Arrays.copyOfRange(concatArray, 0, 50);
				remainingBuffer = Arrays.copyOfRange(concatArray, 50, concatArray.length);
				return retVal;
			}
		};
		WriterMuxSink sink = new WriterMuxSink("4", "test/greek_output.txt",
				"Cp1253");

		// Hook everything up
		transformerA.hookupFaucet(faucet);
		transformerB.hookupFaucet(transformerA);
		transformerB.hookupSink(sink);

		// Start the ball rolling
		sink.consume();

		// Shut things down
		sink.dispose();

		// Verify the result
		Reader inputFileReader = new InputStreamReader(new FileInputStream(
				"test/greek_input.txt"), "UTF-8");
		Reader outputFileReader = new InputStreamReader(new FileInputStream(
				"test/greek_output.txt"), "Cp1253");
		Util.testReaderEquivalence(inputFileReader, outputFileReader);
		inputFileReader.close();
		outputFileReader.close();
		
		File fileToDelete = new File("test/greek_output.txt");
		fileToDelete.delete();
	}

	public void testDemuxStreamTransform() throws IOException {
		// Set up the transformation
		StreamFaucet faucet = new StreamFaucet("1", "test/greek_input.txt");
		IdentityStreamToStreamPipelineItemTemplate transformerB =
			new IdentityStreamToStreamPipelineItemTemplate();
		StreamDemuxPipelineItem transformerA = new StreamDemuxPipelineItem(
				"2", transformerB) {
			private byte[] remainingBuffer;

			@Override
			public boolean shouldChangePipeline() {
				return remainingBuffer != null && remainingBuffer.length != 0;
			}
			
			@Override
			public byte[] transformBuffer(byte[] buffer, int off, int len) {
				if (remainingBuffer == null) {
					if (len == -1)
						return new byte[0];
					
					if (len - off <= 50)
						return Arrays.copyOfRange(buffer, off, off + len);
					byte[] retVal = Arrays.copyOfRange(buffer, off, off + 50);
					remainingBuffer = Arrays.copyOfRange(buffer, off + 50, off + len);
					return retVal;
				}
				
				byte[] concatArray = new byte[remainingBuffer.length + len - off];
				for (int i = 0; i < remainingBuffer.length; i++) {
					concatArray[i] = remainingBuffer[i];
				}
				for (int i = off; i < len; i++) {
					concatArray[remainingBuffer.length + i - off] = buffer[i];
				}
				
				if (concatArray.length <= 50) {
					remainingBuffer = null;
					return Arrays.copyOfRange(concatArray, 0, concatArray.length);
				}
				byte[] retVal = Arrays.copyOfRange(concatArray, 0, 50);
				remainingBuffer = Arrays.copyOfRange(concatArray, 50, concatArray.length);
				return retVal;
			}
		};
		StreamMuxSink sink = new StreamMuxSink("4", "test/greek_output.txt");

		// Hook everything up
		transformerA.hookupFaucet(faucet);
		transformerB.hookupFaucet(transformerA);
		transformerB.hookupSink(sink);

		// Start the ball rolling
		sink.consume();

		// Shut things down
		sink.dispose();

		// Verify the result
		Reader inputFileReader = new InputStreamReader(new FileInputStream(
				"test/greek_input.txt"), "UTF-8");
		Reader outputFileReader = new InputStreamReader(new FileInputStream(
				"test/greek_output.txt"), "UTF-8");
		Util.testReaderEquivalence(inputFileReader, outputFileReader);
		inputFileReader.close();
		outputFileReader.close();
		
		File fileToDelete = new File("test/greek_output.txt");
		fileToDelete.delete();
	}

	public void testZipDemuxStreamTransform() throws IOException {
		// Set up the transformation
		StreamFaucet faucet = new StreamFaucet("1", "test/greek_input.txt");
		IdentityStreamToStreamPipelineItemTemplate transformerB =
			new IdentityStreamToStreamPipelineItemTemplate();
		StreamDemuxPipelineItem transformerA = new StreamDemuxPipelineItem(
				"2", transformerB) {
			private byte[] remainingBuffer;

			@Override
			public boolean shouldChangePipeline() {
				return remainingBuffer != null && remainingBuffer.length != 0;
			}
			
			@Override
			public byte[] transformBuffer(byte[] buffer, int off, int len) {
				if (remainingBuffer == null) {
					if (len == -1)
						return new byte[0];
					
					if (len - off <= 50)
						return Arrays.copyOfRange(buffer, off, off + len);
					byte[] retVal = Arrays.copyOfRange(buffer, off, off + 50);
					remainingBuffer = Arrays.copyOfRange(buffer, off + 50, off + len);
					return retVal;
				}
				
				byte[] concatArray = new byte[remainingBuffer.length + len - off];
				for (int i = 0; i < remainingBuffer.length; i++) {
					concatArray[i] = remainingBuffer[i];
				}
				for (int i = off; i < len; i++) {
					concatArray[remainingBuffer.length + i - off] = buffer[i];
				}
				
				if (concatArray.length <= 50) {
					remainingBuffer = null;
					return Arrays.copyOfRange(concatArray, 0, concatArray.length);
				}
				byte[] retVal = Arrays.copyOfRange(concatArray, 0, 50);
				remainingBuffer = Arrays.copyOfRange(concatArray, 50, concatArray.length);
				return retVal;
			}
		};
		StreamZipMuxSink sink = new StreamZipMuxSink("4", "test/greek_output.zip");

		// Hook everything up
		transformerA.hookupFaucet(faucet);
		transformerB.hookupFaucet(transformerA);
		transformerB.hookupSink(sink);

		// Start the ball rolling
		sink.consume();

		// Shut things down
		sink.dispose();

		// Validate the result
		ZipInputStream zis = new ZipInputStream(new FileInputStream("test/greek_output.zip"));
		for (int i = 0; i < 8; i++) {
			ZipEntry entry = zis.getNextEntry();
			assertNotNull(entry);
			InputStream expectedIS = new FileInputStream(
					"test/validation/MuxDemuxTest/testZipDemuxStreamTransform/entry" + i);
			Util.testStreamEquivalence(zis, expectedIS);
			expectedIS.close();
		}
		assertNull(zis.getNextEntry());
		zis.close();

		File fileToDelete = new File("test/greek_output.zip");
		fileToDelete.delete();
	}
	
	public void testDemuxXMLTransform() throws IOException {
		XMLFaucet faucet = new XMLFaucet("1", "test/demux_input.xml");
		XMLFilterPipelineItem transformerA = new XMLFilterPipelineItem("2", "test/no_xform.xsl");
		XMLFilterPipelineItemTemplate transformerC =
			new XMLFilterPipelineItemTemplate("test/no_xform.xsl");
		XPathXMLMuxFilter muxFilter = new XPathXMLMuxFilter("/root/list/item");
		XMLFilterDemuxPipelineItem transformerB =
			new XMLFilterDemuxPipelineItem("2", muxFilter, transformerC);
		XMLZipMuxSink sink = new XMLZipMuxSink("3", "test/demux_xml.zip");
		
		// Hook everythink up
		transformerA.hookupFaucet(faucet);
		transformerB.hookupFaucet(transformerA);
		transformerB.hookupSink(transformerC);
		transformerC.hookupSink(sink);
		
		// Start the ball rolling
		sink.consume();
		
		// Shut things up
		sink.dispose();
		
		// Validate the result
		ZipInputStream zis = new ZipInputStream(new FileInputStream("test/demux_xml.zip"));
		for (int i = 0; i < 8; i++) {
			ZipEntry entry = zis.getNextEntry();
			assertNotNull(entry);
			InputStream expectedIS = new FileInputStream(
					"test/validation/MuxDemuxTest/testDemuxXMLTransform/entry" + i);
			Util.testStreamEquivalence(zis, expectedIS);
			expectedIS.close();
		}
		assertNull(zis.getNextEntry());
		zis.close();

		File fileToDelete = new File("test/demux_xml.zip");
		fileToDelete.delete();
	}

	public void testAltDemuxXMLTransform() throws IOException {
		XMLFaucet faucet = new XMLFaucet("1", "test/demux_input.xml");
		XMLFilterPipelineItemTemplate transformerC =
			new XMLFilterPipelineItemTemplate("test/no_xform.xsl");
		XPathXMLMuxFilter muxFilter = new XPathXMLMuxFilter("/root/list/item");
		XMLFilterDemuxPipelineItem transformerB =
			new XMLFilterDemuxPipelineItem("2", muxFilter, transformerC);
		XMLZipMuxSink sink = new XMLZipMuxSink("3", "test/demux_xml.zip");
		
		// Hook everythink up
		transformerB.hookupFaucet(faucet);
		transformerB.hookupSink(transformerC);
		transformerC.hookupSink(sink);
		
		// Start the ball rolling
		sink.consume();
		
		// Shut things up
		sink.dispose();
		
		// Validate the result
		ZipInputStream zis = new ZipInputStream(new FileInputStream("test/demux_xml.zip"));
		for (int i = 0; i < 8; i++) {
			ZipEntry entry = zis.getNextEntry();
			assertNotNull(entry);
			InputStream expectedIS = new FileInputStream(
					"test/validation/MuxDemuxTest/testDemuxXMLTransform/entry" + i);
			Util.testStreamEquivalence(zis, expectedIS);
			expectedIS.close();
		}
		assertNull(zis.getNextEntry());
		zis.close();

		File fileToDelete = new File("test/demux_xml.zip");
		fileToDelete.delete();
	}

	public void testStressTestMuxDemux() throws IOException {
		for (int i = 0; i < 100; i++) {
			testDemuxLineTransform();
			testDemuxReaderToWriterTransform();
			testDemuxStreamTransform();
			testZipDemuxStreamTransform();
			testDemuxXMLTransform();
			testAltDemuxXMLTransform();
		}
	}

}
