/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.pipeline;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Future;
import java.util.concurrent.atomic.AtomicReference;

import org.apache.log4j.Logger;

import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.IDemuxFaucet;
import com.koutra.dist.proc.model.IFaucet;
import com.koutra.dist.proc.model.IFaucetTemplate;
import com.koutra.dist.proc.model.IMuxSink;
import com.koutra.dist.proc.model.IPipelineItem;
import com.koutra.dist.proc.model.ISimplePipelineItem;
import com.koutra.dist.proc.model.ISink;
import com.koutra.dist.proc.model.ISinkTemplate;
import com.koutra.dist.proc.model.ISplitPipelineItem;
import com.koutra.dist.proc.model.XformationException;
import com.koutra.dist.proc.util.ByteArrayUtil;
import com.koutra.dist.proc.util.ExecutorsHelper;

public abstract class StreamDemuxPipelineItem
		implements ISimplePipelineItem, IDemuxFaucet {
	private static final Logger logger = Logger.getLogger(StreamDemuxPipelineItem.class);
	
	protected String id;
	protected boolean hookedUpFaucet;
	protected boolean hookedUpSink;
	protected ISinkTemplate pipelineTemplate;
	protected IFaucet faucet;
	protected AtomicReference<ISink> sink;
	protected InputStream inputStream;
	protected PipedInputStream inputStreamForFaucet;
	protected List<PipedInputStream> inputStreamListForDisposal;
	protected PipedOutputStream outputStream;
	
	/**
	 * @deprecated Use any of the initializing constructors instead.
	 */
	public StreamDemuxPipelineItem() { }
	
	public StreamDemuxPipelineItem(String id, ISinkTemplate pipelineTemplate) {
		if (!pipelineTemplate.supportsInput(ContentType.ByteStream))
			throw new IllegalArgumentException("SinkTemplate '" +
					pipelineTemplate.getId() +
					"' must support the ByteStream content type.");
		this.id = id;
		this.hookedUpFaucet = false;
		this.hookedUpSink = false;
		this.pipelineTemplate = pipelineTemplate;
		this.sink = new AtomicReference<ISink>(null);
		this.faucet = null;
		this.inputStream = null;
		this.inputStreamForFaucet = null;
		this.inputStreamListForDisposal = new ArrayList<PipedInputStream>();
		this.outputStream = null;
	}

	@Override
	public String getId() {
		return id;
	}

	@Override
	public ISink getSink() {
		return sink.get();
	}

	@Override
	public IFaucet getFaucet() {
		return faucet;
	}

	@Override
	public boolean supportsInput(ContentType contentType) {
		switch (contentType) {
		case ByteStream:
			return true;
		case CharStream:
		case XML:
		case ResultSet:
		default:
			return false;
		}
	}

	@Override
	public boolean supportsOutput(ContentType contentType) {
		switch (contentType) {
		case ByteStream:
			return true;
		case CharStream:
		case XML:
		case ResultSet:
		default:
			return false;
		}
	}

	protected void checkFaucetValidity(IFaucet faucet) {
		if (!faucet.supportsOutput(ContentType.ByteStream))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' must support the ByteStream content type.");
		if (faucet instanceof IDemuxFaucet)
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is an IDemuxFaucet instance. This is not allowed.");
		if (faucet instanceof IFaucetTemplate)
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is an IFaucetTemplate instance. This is not allowed.");
	}

	@Override
	public void hookupFaucet(IFaucet faucet) {
		checkFaucetValidity(faucet);
		if (hookedUpFaucet && this.faucet == faucet) {
			// Effectively a NO-OP.
			return;
		}
		if (hookedUpFaucet)
			throw new XformationException("Trying to hook up an already hooked up faucet.");
		
		this.faucet = faucet;
		this.hookedUpFaucet = true;
		
		this.faucet.hookupSink(this);
	}

	protected void checkSinkValidity(ISink sink) {
		if (!sink.supportsInput(ContentType.ByteStream))
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' must support the ByteStream content type.");
		if (!(sink instanceof IMuxSink) && !(sink instanceof ISinkTemplate))
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' is neither an IMuxSink or an ISinkTemplate instance. " +
					"This is not allowed.");
	}
	
	@Override
	public void hookupSink(ISink sink) {
		checkSinkValidity(sink);
		if (hookedUpSink && this.sink.get() == sink) {
			// Effectively a NO-OP.
			return;
		}
		
		this.sink.set(sink);
		this.hookedUpSink = true;
		
		this.sink.get().hookupFaucet(this);
	}
	
	@Override
	public Object getSource(ContentType contentType) {
		switch (contentType) {
		case CharStream:
		case XML:
		case ResultSet:
			throw new XformationException("Content type: " + contentType +
			" is not supported.");
		}
		
		if (inputStreamForFaucet != null)
			return inputStreamForFaucet;
		
		this.outputStream = new PipedOutputStream();
		try {
			this.inputStreamForFaucet = new PipedInputStream(this.outputStream);
		} catch (IOException e) {
			throw new XformationException("Unable to create hooked piped streams", e);
		}
		this.inputStreamListForDisposal.add(this.inputStreamForFaucet);
		
		if (logger.isTraceEnabled())
			logger.trace("Created input stream " + inputStreamForFaucet + " and output stream " + outputStream);

		return inputStreamForFaucet;
	}
	
	@Override
	public void registerSource(Object source) {
		// Only used in mux/demux cases.
	}

	@Override
	public void handleMuxSwitch() {
		// Only used in mux/demux cases.
	}
	
	@Override
	public void endMux() {
		// Only used in mux/demux cases.
	}

	@Override
	public void dispose() {
		ISink currentSink = sink.get();
		if (currentSink == null)
			return;
		if (currentSink instanceof ISinkTemplate)
			if (((ISinkTemplate) currentSink).isClone())
				return;
		
		try {
			for (PipedInputStream inputStream : inputStreamListForDisposal) {
				inputStream.close();
			}
		} catch (IOException e) {
			throw new XformationException("Unable to close input stream", e);
		}
		
		faucet.dispose();
	}
	
	protected void createPipeline() {
		pipelineTemplate.createClone(this);
	}
	
	@Override
	public String dumpPipeline() {
		return sink.get().dumpPipeline() + "\n" + getClass().getName() + ": " + inputStream +
			"->" + outputStream + "->" + inputStreamForFaucet;
	}
	
	protected void switchPipeline() {
		sink.set(null);
		inputStreamForFaucet = null;
		outputStream = null;
		hookedUpSink = false;
		
		createPipeline();
		
		getSource(ContentType.ByteStream);

		if (sink.get() instanceof IPipelineItem)
			((IPipelineItem) sink.get()).handleMuxSwitch();
		
		if (logger.isTraceEnabled())
			logger.trace("Dumping pipeline:\n" + dumpPipeline());
	}
	
	/**
	 * This method is a callback that performs the transformation. The returned byte
	 * array will get written on the output stream that hooks this pipeline item to its
	 * sink down the pipeline chain.
	 * 
	 * @param buffer The buffer to read bytes from. Only the bytes in the specified
	 * range should be used.
	 * @param off The offset of the first byte in the buffer that is part of the range
	 * to use.
	 * @param len The number of bytes of the valid range, or null if the end of the
	 * input has been reached.
	 * @return The byte array to pass down the pipeline chain.
	 */
	public abstract byte[] transformBuffer(byte[] buffer, int off, int len);

	/**
	 * This method should return true, if the rest of the pipeline should "break"
	 * at this point in the transformation.
	 * 
	 * @return True iff the transformation pipeline should "break" at this point.
	 */
	public abstract boolean shouldChangePipeline();
	
	@Override
	public void consume() {
		throw new XformationException("Calling ISink.consume() on a IPipelineItem " +
				"implementation. Call IPipelineItem.consume(ISink) instead.");
	}

	@Override
	public Future<?> consumeAsynchronously() {
		throw new XformationException("Calling ISink.consumeAsynchronously() on a IPipelineItem " +
				"implementation. Call IPipelineItem.consume(ISink) instead.");
	}

	@Override
	public Object consume(ISink previousSink) {
		if (!hookedUpFaucet)
			throw new XformationException("Pipeline item has not been set up correctly:" +
					"faucet has not been set");
		
		if (previousSink instanceof ISinkTemplate) {
			// If the sink is a clone we must stop the consumption chain here, as the
			// rest of the chain has already been notified the first time around.
			if (((ISinkTemplate) previousSink).isClone())
				return null;
		}
		
	    if (faucet instanceof IPipelineItem)
	    	((IPipelineItem) faucet).consume(this);
	    
		createPipeline();

		inputStream = (InputStream) faucet.getSource(ContentType.ByteStream);

		getSource(ContentType.ByteStream);
		
		if (sink.get() instanceof IPipelineItem)
			((IPipelineItem) sink.get()).handleMuxSwitch();
		
		final ISink fPreviousSink = previousSink; // needed in order to be used below
		
		ExecutorsHelper.getInstance().executeInProc(new Runnable() {
			@Override
			public void run() {
				try {
					int count;
					byte[] buffer = new byte[8 * 1024];
					while ((count = inputStream.read(buffer)) != -1) {
						byte[] transformation = transformBuffer(buffer, 0, count);
						if (logger.isTraceEnabled()) {
							String readInput = ByteArrayUtil.getHexDump(buffer, 0, count);
							String transformedInput = ByteArrayUtil.getHexDump(
									transformation, 0, transformation.length);
							logger.trace("Read " + readInput + " from input stream " +
									inputStream + ", writing " + transformedInput +
									" to output stream " + outputStream);
						}
						outputStream.write(transformation, 0, transformation.length);
						while (shouldChangePipeline()) {
							outputStream.close();
							if (logger.isTraceEnabled())
								logger.trace("About to switch pipeline");
							switchPipeline();
							transformation = transformBuffer(buffer, 0, 0);
							if (logger.isTraceEnabled()) {
								String transformedInput = ByteArrayUtil.getHexDump(
										transformation, 0, transformation.length);
								logger.trace("Writing additional '" +
										transformedInput + "' to output stream " + outputStream);
							}
							outputStream.write(transformation, 0, transformation.length);
						}
					}
					byte[] transformation = transformBuffer(buffer, 0, count);
					if (logger.isTraceEnabled()) {
						String transformedInput = ByteArrayUtil.getHexDump(
								transformation, 0, transformation.length);
						logger.trace("Writing final '" +
								transformedInput + "' to output stream " + outputStream);
					}
					outputStream.write(transformation, 0, transformation.length);
					outputStream.close();
					
					// Send a null down to the mux sink so that it knows that the ball
					// has stopped rolling.
					if (fPreviousSink instanceof IPipelineItem)
						((IPipelineItem) sink.get()).endMux();
					
					logger.debug("Executable for the stream demux pipeline will now exit.");
				} catch (Throwable t) {
					logger.error("Error while running pipeline", t);
				}
			}
		});
		
		return null;
	}

	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		this.id = in.readUTF();
		this.hookedUpFaucet = false;
		this.hookedUpSink = false;
		this.sink = new AtomicReference<ISink>(null);
		this.faucet = null;
		this.inputStream = null;
		this.inputStreamForFaucet = null;
		this.inputStreamListForDisposal = new ArrayList<PipedInputStream>();
		this.outputStream = null;

		String sinkClassName = in.readUTF();
		Class<?> sinkClass;
		try {
			sinkClass = Class.forName(sinkClassName);
		} catch (ClassNotFoundException e) {
			throw new InstantiationException("Unable to load class: " + sinkClassName);
		}
		this.pipelineTemplate = (ISinkTemplate) sinkClass.newInstance();
		this.pipelineTemplate.readFrom(in);
		this.pipelineTemplate.hookupFaucet(this);
	}
	
	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		out.writeUTF(id);
		out.writeUTF(pipelineTemplate.getClass().getCanonicalName());
		pipelineTemplate.writeTo(out);
	}

	@Override
	public List<ISink> getTerminalSinks() {
		if (pipelineTemplate instanceof IFaucet)
			return ((IFaucet) pipelineTemplate).getTerminalSinks();
		else if (pipelineTemplate instanceof ISplitPipelineItem)
			return ((ISplitPipelineItem) pipelineTemplate).getTerminalSinks();
		else {
			List<ISink> retVal = new ArrayList<ISink>();
			retVal.add(pipelineTemplate);
			return retVal;
		}
	}

}
