/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.pipeline;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.InetAddress;

import org.apache.log4j.Logger;

import com.koutra.dist.proc.cluster.PipedInputStream;
import com.koutra.dist.proc.cluster.PipedOutputStream;
import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.FaucetAdapter;
import com.koutra.dist.proc.model.IDemuxFaucet;
import com.koutra.dist.proc.model.IFaucet;
import com.koutra.dist.proc.model.IFaucetTemplate;
import com.koutra.dist.proc.model.IMuxSink;
import com.koutra.dist.proc.model.IPipelineItem;
import com.koutra.dist.proc.model.ISink;
import com.koutra.dist.proc.model.ISinkTemplate;
import com.koutra.dist.proc.model.XformationException;
import com.koutra.dist.proc.util.ByteArrayUtil;
import com.koutra.dist.proc.util.EmptyInputStream;

public class StreamTaskBoundaryTemplate extends AbstractTaskBoundaryTemplate {
	private static final Logger logger = Logger.getLogger(StreamTaskBoundaryTemplate.class);
	
	protected InputStream inputStream;
	protected PipedInputStream inputStreamForFaucet;
	protected PipedOutputStream outputStream;
	
	/**
	 * @deprecated Use any of the initializing constructors instead.
	 */
	public StreamTaskBoundaryTemplate() {
		this.inputStream = null;
		this.inputStreamForFaucet = null;
		this.outputStream = null;
	}
	
	public StreamTaskBoundaryTemplate(String id) {
		super(id);
		this.inputStream = null;
		this.inputStreamForFaucet = null;
		this.outputStream = null;
	}

	@Override
	protected ContentType getSupportedFaucetContentType() {
		return ContentType.ByteStream;
	}

	@Override
	public boolean supportsInput(ContentType contentType) {
		switch (contentType) {
		case ByteStream:
			return true;
		case CharStream:
		case XML:
		case ResultSet:
		default:
			return false;
		}
	}

	@Override
	public boolean supportsOutput(ContentType contentType) {
		switch (contentType) {
		case ByteStream:
			return true;
		case CharStream:
		case XML:
		case ResultSet:
		default:
			return false;
		}
	}

	protected void checkFaucetValidity(IFaucet faucet) {
		if (!faucet.supportsOutput(ContentType.ByteStream))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' must support the ByteStream content type.");
		if (!(faucet instanceof IDemuxFaucet) &&
				!(faucet instanceof IFaucetTemplate) &&
				!(faucet instanceof FaucetAdapter))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is neither an IDemuxFaucet nor an IFaucetTemplate instance, " +
					"nor a FaucetAdapter. " +
					"This is not allowed.");
	}

	protected void checkSinkValidity(ISink sink) {
		if (!sink.supportsInput(ContentType.ByteStream))
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' must support the ByteStream content type.");
		if (!(sink instanceof IMuxSink) && !(sink instanceof ISinkTemplate))
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' is neither an IMuxSink or an ISinkTemplate instance. " +
					"This is not allowed.");
	}

	@Override
	public Object getSource(ContentType contentType) {
		switch (contentType) {
		case CharStream:
		case XML:
		case ResultSet:
			throw new XformationException("Content type: " + contentType +
			" is not supported.");
		}
		
		if (inputStreamForFaucet != null)
			return inputStreamForFaucet;
		try {
			if (isRemoteTaskSide) {
				// We are on the remote end (the cluster task)
				if (sink instanceof IMuxSink) {
					// We are at the end of the pipeline, just before the mux sink
					if (this.outputStream != null)
						return new EmptyInputStream();
					
					this.outputStream = new PipedOutputStream();
					
					if (logger.isTraceEnabled())
						logger.trace("Created output stream (end of pipeline) " +
								outputStream);
					
					return new EmptyInputStream();
				} else { // !(sink instanceof IMuxSink)
					// We are at the start of the pipeline
					this.inputStreamForFaucet = new PipedInputStream(
							InetAddress.getByAddress(remoteServerIP), remoteServerPort);
					
					if (logger.isTraceEnabled())
						logger.trace("Created input stream (start of pipeline) " +
								inputStreamForFaucet + " linked with IP: " +
								InetAddress.getByAddress(remoteServerIP).toString() +
								":" + remoteServerPort);
					
					return inputStreamForFaucet;
				}
			} else { //!isRemoteTaskSide
				// We are on the local end (the pipeline where the mux/demux objects reside)
				if (sink instanceof IMuxSink) {
					// We are at the end of the pipeline, just before the mux sink
					this.inputStreamForFaucet = new PipedInputStream(
							InetAddress.getByAddress(remoteServerIP), remoteServerPort);
					
					if (logger.isTraceEnabled())
						logger.trace("Created input stream (end of pipeline) " +
								inputStreamForFaucet + " linked with IP: " +
								InetAddress.getByAddress(remoteServerIP).toString() +
								":" + remoteServerPort);
					
					return inputStreamForFaucet;
				} else { // !(sink instanceof IMuxSink)
					// We are at the start of the pipeline
					if (this.outputStream != null)
						return new EmptyInputStream();
					
					this.outputStream = new PipedOutputStream();
					
					if (logger.isTraceEnabled())
						logger.trace("Created output stream (start of pipeline) " +
								outputStream);
					
					return new EmptyInputStream();
				}
			}
		} catch (Exception e) {
			throw new XformationException("Unable to get source", e);
		}
	}
	
	@Override
	protected byte[] getOtherEndHostIP() {
		return outputStream.getOtherEndHostIP();
	}
	
	@Override
	protected int getOtherEndPort() {
		return outputStream.getOtherEndPort();
	}
	
	@Override
	public void handleMuxSwitch() {
		if (sink instanceof IPipelineItem) {
			if (!isRemoteTaskSide) {
				try {
					outputStream.connect();
				} catch (IOException e) {
					throw new XformationException(
							"Unable to connect our server socket to a client.", e);
				}
			}
		} else {
			if (!isRemoteTaskSide) {
				sink.registerSource(getSource(ContentType.ByteStream));
			}
		}
		super.handleMuxSwitch();
	}
	
	@Override
	public String dumpPipeline() {
		return sink.dumpPipeline() + "\n" + getClass().getName() + ": " + inputStream +
			"->" + outputStream + "->" + inputStreamForFaucet;
	}
	
	@Override
	public void dispose() {
		try {
			if (inputStreamForFaucet != null)
				inputStreamForFaucet.close();
		} catch (IOException e) {
			throw new XformationException("Unable to close input stream", e);
		}
		
		faucet.dispose();
	}
	
	@Override
	protected byte[] getServerHostIP() {
		return outputStream.getServerHostIP();
	}
	
	@Override
	protected int getServerPort() {
		return outputStream.getServerPort();
	}
	
	@Override
	protected void localStartBoundaryImpl(Object inputSource) {
		inputStream = (InputStream) faucet.getSource(ContentType.ByteStream);
	    
	    getSource(ContentType.ByteStream);

		try {
			int count;
			byte[] buffer = new byte[8 * 1024];
			if (logger.isTraceEnabled())
				logger.trace("Using input stream " + inputStream +
						" for the stream transformation (start of pipeline).");
			while ((count = inputStream.read(buffer)) != -1) {
				String readInput = ByteArrayUtil.getHexDump(buffer, 0, count);
				if (logger.isTraceEnabled())
					logger.trace("Read " + readInput + " from input stream " +
							inputStream + ", writing to output stream " +
							outputStream);
				outputStream.write(buffer, 0, count);
			}
			outputStream.close();
		} catch (Throwable t) {
			logger.error("Error during pipeline thread execution.", t);
		}
	}
	
	@Override
	protected void remoteEndBoundaryImpl(Object inputSource) {
		try {
			outputStream.connect();
		} catch (IOException e) {
			throw new XformationException(
					"Unable to connect our server socket to a client.", e);
		}
		
		inputStream = (InputStream) faucet.getSource(ContentType.ByteStream);
	    
	    getSource(ContentType.ByteStream);

		try {
			int count;
			byte[] buffer = new byte[8 * 1024];
			if (logger.isTraceEnabled())
				logger.trace("Using input stream " + inputStream +
						" for the stream transformation (start of pipeline).");
			while ((count = inputStream.read(buffer)) != -1) {
				String readInput = ByteArrayUtil.getHexDump(buffer, 0, count);
				if (logger.isTraceEnabled())
					logger.trace("Read " + readInput + " from input stream " +
							inputStream + ", writing to output stream " +
							outputStream);
				outputStream.write(buffer, 0, count);
			}
			outputStream.close();
		} catch (Throwable t) {
			logger.error("Error during pipeline thread execution.", t);
		}
	}
	
	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		super.readFrom(in);
		this.inputStream = null;
		this.inputStreamForFaucet = null;
		this.outputStream = null;
		this.isRemoteTaskSide = true;
		boolean serializedRemoteDetailsExist = in.readBoolean();
		if (serializedRemoteDetailsExist) {
			int byteCount = in.readInt();
			this.remoteServerIP = new byte[byteCount];
			in.readFully(this.remoteServerIP);
			this.remoteServerPort = in.readInt();
		}
	}
	
	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		super.writeTo(out);
		boolean serializeRemoteDetails = !(sink instanceof IMuxSink) && outputStream != null;
		out.writeBoolean(serializeRemoteDetails);
		if (serializeRemoteDetails) {
			out.writeInt(outputStream.getServerHostIP().length);
			out.write(outputStream.getServerHostIP());
			out.writeInt(outputStream.getServerPort());
		}
	}
}
