/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.pipeline;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;

import org.apache.log4j.Logger;
import org.xml.sax.InputSource;
import org.xml.sax.XMLFilter;
import org.xml.sax.XMLReader;

import com.koutra.dist.proc.cluster.PipedInputXMLFilter;
import com.koutra.dist.proc.cluster.PipedOutputXMLFilter;
import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.FaucetAdapter;
import com.koutra.dist.proc.model.IDemuxFaucet;
import com.koutra.dist.proc.model.IFaucet;
import com.koutra.dist.proc.model.IFaucetTemplate;
import com.koutra.dist.proc.model.IMuxSink;
import com.koutra.dist.proc.model.IPipelineItem;
import com.koutra.dist.proc.model.ISink;
import com.koutra.dist.proc.model.ISinkTemplate;
import com.koutra.dist.proc.model.XformationException;
import com.koutra.dist.proc.util.EmptyXMLFilter;

public class XMLTaskBoundaryTemplate extends AbstractTaskBoundaryTemplate {
	private static final Logger logger = Logger.getLogger(XMLTaskBoundaryTemplate.class);
	
	protected XMLFilter filter;
	protected PipedInputXMLFilter filterForFaucet;
	protected PipedOutputXMLFilter outputFilter;
	
	/**
	 * @deprecated Use any of the initializing constructors instead.
	 */
	public XMLTaskBoundaryTemplate() {
		this.filter = null;
		this.filterForFaucet = null;
		this.outputFilter = null;
	}
	
	public XMLTaskBoundaryTemplate(String id) {
		super(id);
		this.filter = null;
		this.filterForFaucet = null;
		this.outputFilter = null;
	}

	@Override
	protected ContentType getSupportedFaucetContentType() {
		return ContentType.XML;
	}

	@Override
	public boolean supportsInput(ContentType contentType) {
		switch (contentType) {
		case XML:
			return true;
		case ByteStream:
		case CharStream:
		case ResultSet:
		default:
			return false;
		}
	}

	@Override
	public boolean supportsOutput(ContentType contentType) {
		switch (contentType) {
		case XML:
			return true;
		case ByteStream:
		case CharStream:
		case ResultSet:
		default:
			return false;
		}
	}

	protected void checkFaucetValidity(IFaucet faucet) {
		if (!faucet.supportsOutput(ContentType.XML))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' must support the XML content type.");
		if (!(faucet instanceof IDemuxFaucet) &&
				!(faucet instanceof IFaucetTemplate) &&
				!(faucet instanceof FaucetAdapter))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is neither an IDemuxFaucet nor an IFaucetTemplate instance, " +
					"nor a FaucetAdapter. " +
					"This is not allowed.");
	}

	protected void checkSinkValidity(ISink sink) {
		if (!sink.supportsInput(ContentType.XML))
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' must support the XML content type.");
		if (!(sink instanceof IMuxSink) && !(sink instanceof ISinkTemplate))
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' is neither an IMuxSink or an ISinkTemplate instance. " +
					"This is not allowed.");
	}

	@Override
	public Object getSource(ContentType contentType) {
		switch (contentType) {
		case ByteStream:
		case CharStream:
		case ResultSet:
			throw new XformationException("Content type: " + contentType +
			" is not supported.");
		}
		
		if (filterForFaucet != null)
			return filterForFaucet;
		
		try {
			if (isRemoteTaskSide) {
				// We are on the remote end (the cluster task)
				if (sink instanceof IMuxSink) {
					// We are at the end of the pipeline, just before the mux sink
					if (this.outputFilter != null)
						return new EmptyXMLFilter();
					
					this.outputFilter = new PipedOutputXMLFilter();
					
					if (logger.isTraceEnabled())
						logger.trace("Created output XML filter (end of pipeline) " +
								outputFilter);
					
					return new EmptyXMLFilter();
				} else { // !(sink instanceof IMuxSink)
					// We are at the start of the pipeline
					this.filterForFaucet = new PipedInputXMLFilter(
							InetAddress.getByAddress(remoteServerIP), remoteServerPort);
					
					if (logger.isTraceEnabled())
						logger.trace("Created input XML filter (start of pipeline) " +
								filterForFaucet + " linked with IP: " +
								InetAddress.getByAddress(remoteServerIP).toString() +
								":" + remoteServerPort);
					
					return filterForFaucet;
				}
			} else { //!isRemoteTaskSide
				// We are on the local end (the pipeline where the mux/demux objects reside)
				if (sink instanceof IMuxSink) {
					// We are at the end of the pipeline, just before the mux sink
					this.filterForFaucet = new PipedInputXMLFilter(
							InetAddress.getByAddress(remoteServerIP), remoteServerPort);
					
					if (logger.isTraceEnabled())
						logger.trace("Created input XML filter (end of pipeline) " +
								filterForFaucet + " linked with IP: " +
								InetAddress.getByAddress(remoteServerIP).toString() +
								":" + remoteServerPort);
					
					return filterForFaucet;
				} else { // !(sink instanceof IMuxSink)
					// We are at the start of the pipeline
					if (this.outputFilter != null)
						return new EmptyXMLFilter();
					
					this.outputFilter = new PipedOutputXMLFilter();
					
					if (logger.isTraceEnabled())
						logger.trace("Created output XML filter (start of pipeline) " +
								outputFilter);
					
					return new EmptyXMLFilter();
				}
			}
		} catch (Exception e) {
			throw new XformationException("Unable to get source", e);
		}
	}
	
	@Override
	protected byte[] getOtherEndHostIP() {
		return outputFilter.getOtherEndHostIP();
	}
	
	@Override
	protected int getOtherEndPort() {
		return outputFilter.getOtherEndPort();
	}
	
	@Override
	public void handleMuxSwitch() {
		if (sink instanceof IPipelineItem) {
			if (!isRemoteTaskSide) {
				try {
					outputFilter.connect();
				} catch (IOException e) {
					throw new XformationException(
							"Unable to connect our server socket to a client.", e);
				}
			}
		} else {
			if (!isRemoteTaskSide) {
				sink.registerSource(this);
			}
		}
		super.handleMuxSwitch();
	}
	
	@Override
	public String dumpPipeline() {
		return sink.dumpPipeline() + "\n" + getClass().getName() + ": " + filter +
			"->" + outputFilter + "->" + filterForFaucet;
	}
	
	@Override
	public void dispose() {
		try {
			if (filterForFaucet != null)
				filterForFaucet.close();
		} catch (IOException e) {
			logger.warn("Unable to close reader" + e.getMessage());
			// eat the exception up.
		}
		
		faucet.dispose();
	}
	
	@Override
	protected byte[] getServerHostIP() {
		return outputFilter.getServerHostIP();
	}
	
	@Override
	protected int getServerPort() {
		return outputFilter.getServerPort();
	}
	
	@Override
	protected void localStartBoundaryImpl(Object inputSource) {
		outputFilter.setParent((XMLReader) faucet.getSource(ContentType.XML));
	    
	    getSource(ContentType.XML);

		try {
			outputFilter.drivePipeline((InputSource) inputSource);
		} catch (Throwable t) {
			logger.error("Error during pipeline thread execution.", t);
		}
	}
	
	@Override
	protected void remoteEndBoundaryImpl(Object inputSource) {
		try {
			outputFilter.connect();
		} catch (IOException e) {
			throw new XformationException(
					"Unable to connect our server socket to a client.", e);
		}
		
	    getSource(ContentType.XML);

		filter = (XMLFilter) faucet.getSource(ContentType.XML);
		outputFilter.setParent(filter);
	    
		try {
			outputFilter.drivePipeline((InputSource) inputSource);
		} catch (Throwable t) {
			logger.error("Error during boundary to boundary execution.", t);
		}
	}
	
	@Override
	protected Object initialInputSource() {
		return new InputSource();
	}
	
	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		super.readFrom(in);
		this.filter = null;
		this.filterForFaucet = null;
		this.outputFilter = null;
		this.isRemoteTaskSide = true;
		boolean serializedRemoteDetailsExist = in.readBoolean();
		if (serializedRemoteDetailsExist) {
			int byteCount = in.readInt();
			this.remoteServerIP = new byte[byteCount];
			in.readFully(this.remoteServerIP);
			this.remoteServerPort = in.readInt();
		}
	}
	
	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		super.writeTo(out);
		boolean serializeRemoteDetails = !(sink instanceof IMuxSink) && outputFilter != null;
		out.writeBoolean(serializeRemoteDetails);
		if (serializeRemoteDetails) {
			out.writeInt(outputFilter.getServerHostIP().length);
			out.write(outputFilter.getServerHostIP());
			out.writeInt(outputFilter.getServerPort());
		}
	}
}
