/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.sink;

import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.concurrent.Future;

import javax.xml.transform.Result;
import javax.xml.transform.Transformer;
import javax.xml.transform.sax.SAXResult;
import javax.xml.transform.sax.SAXSource;

import org.apache.fop.apps.FOUserAgent;
import org.apache.fop.apps.Fop;
import org.apache.fop.apps.FopFactory;
import org.apache.log4j.Logger;
import org.apache.xalan.transformer.TrAXFilter;
import org.xml.sax.InputSource;
import org.xml.sax.XMLFilter;
import org.xml.sax.XMLReader;

import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.IDemuxFaucet;
import com.koutra.dist.proc.model.IFaucet;
import com.koutra.dist.proc.model.IFaucetTemplate;
import com.koutra.dist.proc.model.IPipelineItem;
import com.koutra.dist.proc.model.ISink;
import com.koutra.dist.proc.model.XformationException;
import com.koutra.dist.proc.util.ExecutorsHelper;

public class FOPSink implements ISink {
	private static final Logger logger = Logger.getLogger(FOPSink.class);
	
	protected enum Type {
		File,
		OutputStream;
	};
	
	protected String id;
	protected FopFactory fopFactory;
	protected boolean hookedUp;
	protected IFaucet faucet;
	protected Object faucetObject;
	protected String fopTargetMimeType;
	protected Type type;
	protected OutputStream outputStream;
	protected String path;
	
	/**
	 * @deprecated Use any of the initializing constructors instead.
	 */
	public FOPSink() { }
	
	public FOPSink(String id, String fopTargetMimeType, String path) {
		this(id, FopFactory.newInstance(), fopTargetMimeType, path);
	}
	
	public FOPSink(String id, FopFactory fopFactory, String fopTargetMimeType,
			String path) {
		this.id = id;
		this.fopFactory = fopFactory;
		this.hookedUp = false;
		this.faucet = null;
		this.fopTargetMimeType = fopTargetMimeType;
		this.type = Type.File;
		this.outputStream = null;
		this.path = path;
	}

	public FOPSink(String id, String fopTargetMimeType, OutputStream outputStream) {
		this(id, FopFactory.newInstance(), fopTargetMimeType, outputStream);
	}
	
	public FOPSink(String id, FopFactory fopFactory, String fopTargetMimeType,
			OutputStream outputStream) {
		this.id = id;
		this.fopFactory = fopFactory;
		this.hookedUp = false;
		this.faucet = null;
		this.fopTargetMimeType = fopTargetMimeType;
		this.type = Type.OutputStream;
		this.outputStream = outputStream;
		this.path = null;
	}

	@Override
	public String getId() {
		return id;
	}

	@Override
	public IFaucet getFaucet() {
		return faucet;
	}

	@Override
	public boolean supportsInput(ContentType contentType) {
		switch (contentType) {
		case XML:
			return true;
		case ByteStream:
		case CharStream:
		case ResultSet:
		default:
			return false;
		}
	}

	@Override
	public void hookupFaucet(IFaucet faucet) {
		if (!faucet.supportsOutput(ContentType.XML))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' must support the XML content type.");
		if (faucet instanceof IDemuxFaucet)
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is an IDemuxFaucet instance. This is not allowed.");
		if (faucet instanceof IFaucetTemplate)
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is an IFaucetTemplate instance. This is not allowed.");
		if (hookedUp && this.faucet == faucet) {
			// Effectively a NO-OP.
			return;
		}
		if (hookedUp)
			throw new XformationException("Trying to hook up an already hooked up sink.");
		
		this.faucet = faucet;
		this.hookedUp = true;
		
		this.faucet.hookupSink(this);
	}

	@Override
	public void registerSource(Object source) {
		// Only used in mux/demux situations.
	}
	
	@Override
	public String dumpPipeline() {
		return getClass().getName() + ": " + faucetObject;
	}

	@Override
	public void dispose() {
		switch (type) {
		case File:
			try {
				outputStream.close();
			} catch (IOException e) {
				throw new XformationException("Unable to close output stream", e);
			}
			break;
		case OutputStream:
			// Caller owns the output stream. It is their responsibility to close it.
			break;
		}
		
		faucet.dispose();
	}

	private XMLReader getRealReader(Object previousFilter) {
		if (previousFilter instanceof XMLFilter) {
			XMLFilter filter = (XMLFilter) previousFilter;
			return getRealReader(filter.getParent());
		} else {
			return (XMLReader) previousFilter;
		}
	}
	
	@Override
	public void consume() {
		if (!hookedUp)
			throw new XformationException("Sink has not been set up correctly: " +
					"faucet has not been set");
		
		switch (type) {
		case File:
			try {
				outputStream = new BufferedOutputStream(new FileOutputStream(path));
			} catch (FileNotFoundException e) {
				throw new XformationException("Unable to create output stream", e);
			}
			break;
		case OutputStream:
			break;
		}
		
	    InputSource input = null;
	    if (faucet instanceof IPipelineItem) {
	    	input = (InputSource) ((IPipelineItem) faucet).consume(this);
	    }

		FOUserAgent foUserAgent = fopFactory.newFOUserAgent();
		
		try {
			faucetObject = faucet.getSource(ContentType.XML);
			if (logger.isTraceEnabled())
				logger.trace("Sink is using reader: " + faucetObject);
			XMLFilter filter = (XMLFilter) faucetObject;
			if (!(filter instanceof TrAXFilter))
				throw new XformationException("Unsupported transformation class " +
						filter.getClass().getName());
		    Transformer transformer = ((TrAXFilter) filter).getTransformer();
			Fop fop = fopFactory.newFop(fopTargetMimeType, foUserAgent, outputStream);
			Result result = new SAXResult(fop.getDefaultHandler());
		    SAXSource transformSource = new SAXSource(getRealReader(filter), input);
		    transformer.transform(transformSource, result);
		} catch (Exception e) {
			throw new XformationException("Unable to set up FOP", e);
		}
	}
	
	@Override
	public Future<?> consumeAsynchronously() {
		return ExecutorsHelper.getInstance().submitInProc(new Runnable() {
			@Override
			public void run() {
				try {
					consume();
				} catch (Throwable t) {
					logger.error("Unable to consume faucet", t);
				}
			}
		});
	}

	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		this.id = in.readUTF();
		this.fopFactory = FopFactory.newInstance();
		this.hookedUp = false;
		this.faucet = null;
		this.fopTargetMimeType = in.readUTF();
		this.type = Type.File;
		this.outputStream = null;
		this.path = in.readUTF();
	}

	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		if (type != Type.File)
			throw new XformationException(
					"Only FOPSinks of type Type.File can be streamed.");
		
		out.writeUTF(id);
		out.writeUTF(fopTargetMimeType);
		out.writeUTF(path);
	}

}
